/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  This file is part of the ZipCracker Program.

    Copyright (C) 1998-2000 Jonas Borgstrm <jonas_b@bitsmart.com>.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/
#include <gtk/gtk.h>
#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <zlib.h>
#include "crack.h"
#include "crc_table.h"
#include "zip.h"
#include "zipcracker.h"
/*****************************************************************************/
extern Properties *prop;
extern gchar *fileformat_magic;
extern gint fileformat_magic_len;
/*****************************************************************************/
static char *format_passwd (const char *guess, char *realpw);
static void update_keys (unsigned long *key, char val);
static unsigned char decrypt_byte (const unsigned long *key);
static int decript_head (unsigned char *buffer, unsigned long *key);
static unsigned long Crc32 (unsigned long oldcrc, gchar newchar);
static gboolean test_password (unsigned long *key);
static void crack ();
/*****************************************************************************/
static FILE *zipfd = NULL;
static gchar passwd[PW_LEN];
static gchar realpw[PW_LEN]; /* Passwd after running format_passwd */
static gchar *format = NULL;
static guchar header_crc;
static guchar *header;
static guint8 method;
static gulong file_offset;
static gulong orig_size, comp_size;
static gulong CRC32;
static gint crack_is_running = 0;
static gint flag_found_password;
static gint crack_count;
/*****************************************************************************/
/*   Returns 1 if the thread is cracking */
/*****************************************************************************/
gint crack_is_cracking()
{
	return crack_is_running;
}
/*****************************************************************************/
/*   Stop the thread */
/*****************************************************************************/
void
crack_stop_cracking()
{
	crack_is_running = 0;
}
/*****************************************************************************/
/*   Returns the current password */
/*****************************************************************************/
gchar *
crack_get_current_password() 
{
	return realpw;
}
/*****************************************************************************/
/*   Returns the current raw password */
/*****************************************************************************/
gchar *
crack_get_current_raw_password() 
{
	return passwd;
}
/*****************************************************************************/
/*   Returns true if we found the correct password */
/*****************************************************************************/
gint
crack_found_correct_password()
{
	return flag_found_password;
}
/*****************************************************************************/
/*   Returns the current speed */
/*****************************************************************************/
gint
crack_get_count()
{
	gint tmp = crack_count;
	crack_count = 0;
	return tmp;
}
/*****************************************************************************/
/*   Initializes the password and checks for errors */
/*****************************************************************************/
gint
crack_init_passwd () 
{
	int minlen = 0, len, i;
	guint8 tmp[256];
	gchar *ptr;
	
	passwd[0] = *prop->password_string->str;
	passwd[1] = 0;

	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(prop->checkbox_startpw))) {
		/* Can't use a start string of 0 length */
		if (strlen (gtk_entry_get_text
			(GTK_ENTRY(prop->entry_startpw))) == 0)
			return 2;
	
		strcpy(passwd, gtk_entry_get_text
			(GTK_ENTRY(prop->entry_startpw)));
	}

	for(i=0;i<strlen(passwd);i++) {
		if(strchr(prop->password_string->str, passwd[i]) == NULL) {

			return 1;
		}
	}
	memset(tmp, 0, 256 * sizeof(guint8));

	i = 0;
	ptr = g_strdup(prop->password_string->str);
	
	/* Removes dublets */	
	while(ptr[i]) {

		if(tmp[(int)ptr[i]])
			memmove(&ptr[i], &ptr[i + 1], strlen(ptr) - i + 1);
		else
			tmp[(int)ptr[i++]]++;
	}
	g_string_assign(prop->password_string, ptr);

	g_free(ptr);

	if (format)
		g_free (format);

	/* Use user specified format string if availible */
	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(prop->checkbox_format))) {

		ptr = gtk_entry_get_text (GTK_ENTRY (prop->entry_format));
		/* The format string must contain an asterix */
		if (strchr (ptr, '*') == NULL) {

			format = g_malloc (strlen (ptr) + 2);
			sprintf (format, "%s*", ptr);
		} else
			format = g_strdup (ptr);

	} else
		format = g_strdup ("*");

	/* Calculating min length of password */
	ptr = format;
	while( (ptr = (strchr(ptr, '?') + 1)) != (char *)1)
		minlen++;

	/* Make it longer if it's to short */
	if(strlen(passwd) < minlen) {
		len = strlen(passwd);
		for(i=len;i<minlen;i++) {
			passwd[i] = prop->password_string->str[0];
		}
		passwd[minlen] = 0;
	}
	flag_found_password = 0;
	crack_count = 0;
	
	return 0;
}
/*****************************************************************************/
/* Entry point of the crack thread */
/*****************************************************************************/
void *
crack_start (void *arg)
{
	ZipFileEntry *target;
	
	crack_is_running = 1;

	target = (ZipFileEntry *)arg;
	
   	header_crc 	= target->header_crc;
	header 		= target->random_header;
	method 		= target->method;
	file_offset	= target->file_offset;  	
	comp_size	= target->comp_size;  	
	orig_size	= target->orig_size;  	
	CRC32		= target->crc32;
	
 	if((zipfd = fopen (prop->zipfilename->str, "rb")) == NULL) {
	   
		g_print ("Can't open '%s'\n", prop->zipfilename->str);
	   
	   	exit (0);
	}
   
   	crack();
	
	crack_is_running = 0;
	
   	fclose(zipfd);

	return NULL;
}
/****************************************************************************/
/*    Main cracking procedure */
/****************************************************************************/
static void 
crack ()
{
	int i, len;
	unsigned long key[3];
	unsigned char hackbuff[19];
   
   	len = strlen(passwd);
	
	while(crack_is_running) {
		/* Used to calculate the speed */
		crack_count++;
		
		format_passwd(passwd, realpw);
		
		memcpy(hackbuff, header, 12);
		key[0] = 305419896L;
		key[1] = 591751049L;
		key[2] = 878082192L;
	   
		for(i=0; i<strlen(realpw);i++)
			update_keys(key,realpw[i]);
	   
		/* Is password a possible password */
		if(decript_head(hackbuff,key)) {

			if(test_password(key)) {
			   
				flag_found_password = 1;
				return;
			}			
		}
	   
		i = len - 1;
	   
		passwd[i] = *(strchr(prop->password_string->str, passwd[i]) + 1);
   		while(passwd[i] == 0) {
		   	if(i == 0) {
				   
			   	memmove(passwd + 1, passwd, len);
			   	passwd[0] = prop->password_string->str[0];
			   	passwd[1] = prop->password_string->str[0];
			   	len++;
			   	passwd[len] = 0;
				break;	
			   
			} else if(passwd[i] == 0) {
				   
				passwd[i] = prop->password_string->str[0];
				passwd[i - 1] = *(strchr(prop->password_string->str, passwd[i - 1]) + 1);
			}
		   	i--;
		   	if(i == -1)
		     		break;
		};
	}
}
/*****************************************************************************/
/*    Tries to decrypt header and checks if CRC match, if so return 1 */
/*****************************************************************************/
static int 
decript_head(unsigned char *buffer, unsigned long *key)
{
	int i;
	char c;
	
	for(i = 0; i<12;i++) {
		c = buffer[i] ^ decrypt_byte(key);
		update_keys(key,c);
		buffer[i] = c;
	}
	
	return buffer[11] == header_crc;
}
/*****************************************************************************/
/*    update_keys() */
/*****************************************************************************/
static void
update_keys(unsigned long *key, char val)
{
	
	key[0] = Crc32(key[0], val);
	key[1] = (key[1] + (key[0] & 0xff)) * 134775813L + 1;
	key[2] = Crc32(key[2], (key[1] >> 24));
}
/*****************************************************************************/
/*    decrypt_byte */
/*****************************************************************************/
static unsigned char decrypt_byte(const unsigned long *key) {
	unsigned short temp;
	
	temp = key[2] | 2;
	return((temp *( temp^1)) >> 8);
}
/*****************************************************************************/
/*    CRC rutine */
/*****************************************************************************/
static unsigned long 
Crc32 (unsigned long oldcrc, gchar newchar)
{
	register unsigned long Crc = oldcrc;
	
        Crc = ((Crc >> 8) & 0x00ffffffL) ^ 
		crc_table[(Crc ^ newchar) & 0xffL];
	
	return Crc;            /* return a 1's complement */
}
/*****************************************************************************/
/*    Test if the password is correct, by unpacking and testing CRC */
/*****************************************************************************/
static gboolean
test_password (unsigned long *key)
{
	gchar inbuffer[REAL_BUF_SIZE];
	gchar outbuffer[REAL_BUF_SIZE];
	int err = Z_OK, i, n;
	int in_buf_size, out_buf_size;
	struct z_stream_s zstrm;
	char c;
	unsigned long newcrc;
	long bl;
   	gboolean check_fileformat_magic;


	check_fileformat_magic = fileformat_magic_len > 0;
   
	out_buf_size = (fileformat_magic_len > 0) ? (fileformat_magic_len) : REAL_BUF_SIZE;
   
	/* Initilize CRC */
	newcrc = crc32(0L, Z_NULL, 0);
		
	/* Check compresion method */
   
	if(method == DEFLATED) {
		/* Use 2x size just to be safe */
		fseek(zipfd, file_offset, SEEK_SET);
		
		zstrm.avail_in = fread(inbuffer, 1, (comp_size < REAL_BUF_SIZE) ? comp_size : REAL_BUF_SIZE, zipfd);
		
		for(i = 0; i<zstrm.avail_in;i++) {

			c = inbuffer[i] ^ decrypt_byte(key);
			update_keys(key,c);
			inbuffer[i] = c;
		}
		
		zstrm.next_out = outbuffer;
		zstrm.avail_out = out_buf_size;
		zstrm.next_in = inbuffer;
		zstrm.zalloc = (alloc_func)Z_NULL;
		zstrm.zfree = (free_func)Z_NULL;
		zstrm.opaque = (free_func)Z_NULL;
		
		/* Initilize zlib */
		if((err = inflateInit2(&zstrm, -15)) != Z_OK ) {
			
			g_error("inflateInit() failed (%d)\n", err);
		}
		
		for(;;) {			
			err = inflate(&zstrm, Z_NO_FLUSH);
			/* Did it go wrong? */
			if (err < 0) {
				
				inflateEnd (&zstrm);
				return FALSE;
			}
			
			if(check_fileformat_magic && zstrm.avail_out == 0) {

				check_fileformat_magic = FALSE;

				if(memcmp(outbuffer, fileformat_magic, fileformat_magic_len) != 0) {
					/* No can't be correct password */
					inflateEnd (&zstrm);
					return FALSE;
				}
				/* Yes might be correct password */
			   	in_buf_size = REAL_BUF_SIZE;
			}
			
			if(zstrm.avail_in == 0) {
				zstrm.avail_in = fread(inbuffer, 1, ((comp_size - zstrm.total_in)< REAL_BUF_SIZE) ? 
						       (comp_size - zstrm.total_in) : REAL_BUF_SIZE, zipfd);

				for(i = 0; i<zstrm.avail_in;i++) {

					c = inbuffer[i] ^ decrypt_byte(key);
					update_keys(key,c);
					inbuffer[i] = c;
				}
				zstrm.next_in = inbuffer;
			}
			/* Run CRC on the output */
			newcrc = crc32(newcrc, outbuffer, out_buf_size - zstrm.avail_out);
			zstrm.next_out = outbuffer;
			zstrm.avail_out = out_buf_size = REAL_BUF_SIZE;
	
			/* Are we finished? */
			if (err == Z_STREAM_END)
				break;			
		}
		inflateEnd (&zstrm);

	} else if (method == STORED) {
		
		in_buf_size = (fileformat_magic_len > 0) ? fileformat_magic_len : REAL_BUF_SIZE;
		fseek(zipfd, file_offset, SEEK_SET);
		bl = orig_size;
		
		while ((n = fread(inbuffer, 1, (bl > in_buf_size) ? in_buf_size : bl, zipfd))) {
			bl -= n;
			/* Decrypt */
			for (i = 0; i<n;i++) {
				c = inbuffer[i] ^ decrypt_byte(key);
				update_keys(key,c);
				inbuffer[i] = c;
			}
		   	if (check_fileformat_magic) {
				check_fileformat_magic = FALSE;
				
			   	if(memcmp(inbuffer, fileformat_magic, fileformat_magic_len) != 0)
			     		return FALSE;

			   	in_buf_size = REAL_BUF_SIZE;
			   	out_buf_size = REAL_BUF_SIZE;
			}
			newcrc = crc32(newcrc, inbuffer, n);
		};
	}
					
	return(CRC32 == newcrc);
}
/****************************************************************************/
/*    Formats the real password */
/****************************************************************************/
static char *
format_passwd (const char *guess, char *realpw) 
{
	int i, j = 0, k = 0;

	for(i=0;i<strlen(format);i++) {
		switch(format[i]) {
		case '?':
			if(guess[k])
				realpw[j++] = guess[k++];
			break;
		case '*':
			while(guess[k] != 0) 
				realpw[j++] = guess[k++];
			break;
		default:
			realpw[j++] = format[i];
		};	
	}	
	realpw[j] = 0;
	
	return realpw;
}
/****************************************************************************/
